!******************************************************************************
!
! File : jd_module.f90
!   by : Alan Tackett
!   on : 02/09/00
!  for : General Purpose
!
!  This file contains routines to implement a block version of the 
!  Jacobi-Davidson QZ solver.  This is based on code by Sleipjen, et al.
!  The calling routine is JD_Solver and is designed as an INNER iteration
!  only.  The user should call JD_solver repeatedly until convergence.
!   
!  The program expects the external routines ABMUL and PRECON to be defined
!  with the following formats:
!
!      Subroutine ABMUL(n, q, Aq, Bq)
!          Use jd_misc
!          Integer :: n
!          Complex :: q(n)
!          Complex :: Aq(n)
!          Complex :: Bq(n)
!
!          --Calculate Aq and Bq and return the result--
!
!          Return
!      End Subroutine 
!
!      Subroutine Precon(n, q)
!          Use jd_misc
!          Integer :: n
!          Complex :: q(n)
!          Complex :: Aq(n)
!          Complex :: Bq(n)
!
!          --Calculate q = K^(-1)q --
!
!          Return
!      End Subroutine 
!
!  Notice that both routines use the module jd_misc.  This module contains
!  the "global" data that needs to be passed to the routine and specifically
!  energy "shift" which is contained in the variable SHIFT_ENERGY.
!
!******************************************************************************

Module jd_module

Use search_sort
Use relax
Use projectors
Use local_criteria_lib
Use hamiltonian


Implicit NONE!!!!!

Integer, PARAMETER :: JD_GMRES  = 1
Integer, PARAMETER :: JD_CGSTAB = 2

Integer, PARAMETER :: ORDER_NEAREST  =  0
Integer, PARAMETER :: ORDER_SMALLEST = -1
Integer, PARAMETER :: ORDER_LARGEST  = -2

Contains

Include "jdqz_interior.f90"

!******************************************************************************
! JD_Sort - reorders the matrix so the wanted eigenvalues are at the front
!******************************************************************************

Subroutine JD_Sort(target, N, alpha, beta, LUT, Order)
  Real,    Intent(IN)    :: Target
  Integer, Intent(IN)    :: N
  Complex, Intent(INOUT) :: alpha(:)
  Complex, Intent(INOUT) :: beta(:)
  Integer, Intent(OUT)   :: LUT(:)
  Integer, Intent(IN)    :: Order

  Integer :: i, j, k
  Real    :: Eigen(N)
  Complex :: temp(N,N), eig(N,2)

  !*** Make the list of eigenvalues ***
  Do i=1, N
     Eigen(i) = alpha(i)/beta(i)
  End do
  
  !*** Adjust list based on ordering ***
  If (Order == ORDER_NEAREST) Eigen = ABS(Eigen-Target)

  !*** Sort them in ascending/descending energy ***
  If (Order == ORDER_LARGEST) then
     Call Insertion_sort(Eigen, LUT, .FALSE.)
  else
     Call Insertion_sort(Eigen, LUT, .TRUE.)
  End If

  Return
End Subroutine

!******************************************************************************
!  JDQZ_ProcessCluster - Generates the clusterinfo so that each cluster
!       is processed individually.
!******************************************************************************

Subroutine JDQZ_ProcessCluster(NBands, ClusterMap, ClusterInfo, Qmax, Qmin)
  Integer, Intent(IN) :: Nbands
  Integer, Intent(INOUT) :: ClusterMap(:)
  Integer, Intent(INOUT) :: ClusterInfo(:,:)
  Integer, Intent(IN) :: Qmax
  Integer, Intent(IN) :: Qmin

  integer :: i, j,s,t, NCluster, cluster, drop, middle

  Cluster = ClusterMap(1)
  NCluster = 0
  drop = 0
  middle = 0
  Do i=1, Nbands
     If (ClusterMap(i) /= Cluster) then
        ClusterInfo(:,cluster) = (/ncluster,drop/)

        drop = middle
        middle = ncluster

        cluster = ClusterMap(i)
        ncluster = 0
     End If
     
     ncluster = ncluster + 1
  End do

  ClusterInfo(:,cluster) = (/ncluster,drop/)

Write(*,*) 'JDQZ_ClusterPROCESS: Final Size=',cluster
j = 0
s=0; t=0;
Do i=1, cluster
   Write(*,*) 'JDQZ_Cluster: ',i, ' * add,drop=',ClusterInfo(:,i)
   
   Write(*,*) '  * OInit=',ClusterMap(s+1:t)
   s = s + ClusterInfo(2,i)
   t = t + ClusterInfo(1,i)
   Write(*,*) '  *  OEnd=',ClusterMap(s+1:t)
   j = j + ClusterInfo(1,i)
End do
Write(*,*) 'JDQZ_Cluster: Final ClusterMap=',ClusterMap(1:nbands)

  Return
End Subroutine


!******************************************************************************
!  JDQZ_Cluster - Reclusters the given "clusters" so that if possible a
!       complete cluster will always be able to be stored in the converged
!       eigenvector storage.
!******************************************************************************

Subroutine JDQZ_cluster(NBands, ClusterMap, ClusterInfo, Qmax, Qmin)
  Integer, Intent(IN) :: Nbands
  Integer, Intent(INOUT) :: ClusterMap(:)
  Integer, Intent(INOUT) :: ClusterInfo(:,:)
  Integer, Intent(IN) :: Qmax
  Integer, Intent(IN) :: Qmin

  integer :: i, j, k, s, t, Qsize, NCluster, Cpos
  Integer :: ClusterSize(Nbands), Cluster, OrgClusters, dropped, kept
  Integer :: ClusterCache(Qmax), CacheSize, OrgSize(Nbands)
  Integer :: OrgMap(Nbands)
  Logical :: Ok, CloseCluster

  !** First determine how many bands are in each cluster **
  OrgSize = 0
  Cluster = ClusterMap(1)
  NCluster = 1
  Do i=1, Nbands
     If (ClusterMap(i) /= Cluster) then
        Ncluster = Ncluster + 1
        cluster = ClusterMap(i)
     End If
        
     OrgSize(Ncluster) = OrgSize(Ncluster) + 1
     ClusterMap(i) = cluster
  End do

  OrgMap(1:Nbands) = ClusterMap(1:nbands)

  !*** Now determine the new clustering ***
  OrgClusters = NCluster
  NCluster = 0
  Qsize = 0
  Cpos = 0
  CacheSize = 0
  ClusterCache = 0
  ClusterInfo = 0
  ClusterSize = 0

Write(*,*) 'JDQZ_Cluster: OrgClusters=',OrgClusters, ' * Qmax=',Qmax, ' * Qmin=',Qmin
Write(*,*) 'JDQZ_Cluster: Start ClusterMap=',OrgMap(1:nbands)
Write(*,*) 'JDQZ_Cluster: Start orgsize=',orgsize(1:orgclusters)

  CloseCluster= .FALSE.
  Do i=1, OrgClusters
Write(*,*) 'i=',i, ' * CS=',OrgSize(i), ' * ccse=',CacheSize, ' * cc=',ClusterCache(1:CacheSize), ' * nc=',ncluster
     k = OrgSize(i)+SUM(ClusterCache(1:CacheSize))

     If ((k <= Qmax) .AND. (i/=OrgClusters)) then  !**Don't have to drop vecs**
Write(*,*) '  SIMPLE'
        CloseCluster = .TRUE.                               !B
        ClusterMap(cpos+1:cpos+OrgSize(i)) = Ncluster+1 !B

        Ncluster = Ncluster + 1                           !A
        CloseCluster = .FALSE.                            !A
        ClusterMap(cpos+1:cpos+OrgSize(i)) = Ncluster !A
        
        CacheSize = CacheSize + 1
        Qsize = Qsize + OrgSize(i)
        ClusterCache(CacheSize) = OrgSize(i)
        cpos = cpos + Orgsize(i)
        ClusterInfo(:,ncluster) = (/OrgSize(i), 0/)
     else if (k > Qmax) then   !** Got to drop some vecs ***
Write(*,*) '  DROP'
        if (CloseCluster) then
           ncluster = ncluster + 1
           ClusterInfo(1,ncluster) = Qsize
           ClusterInfo(2,ncluster) = 0
           CloseCluster = .FALSE.
Write(*,*) '   nc=',ncluster, ' * add,drop=',ClusterInfo(:,ncluster), ' * Close'
        End If

        s = Orgsize(i)

        !*** Determine the dropping point for next cluster ***
        j = 0
        Ok = .FALSE.
        t = -1
        Do While (.NOT. Ok .AND. (j < CacheSize))
           j = j + 1
           k = SUM(ClusterCache(j:CacheSize)) + s
!           If (k<=Qmax) ok = .TRUE.
           If (k<=Qmax) then
              If ((k-s)<Qmin) then
                 Ok = .TRUE.
                 If (j>1) j = j - 1
                 k = SUM(ClusterCache(j:CacheSize)) + s
                 If (k>Qmax) j = j + 1
                 k = SUM(ClusterCache(j:CacheSize)) + s
              End If
           End If
        End do

Write(*,*) '    j=', j, ' * CC=',ClusterCache(1:CacheSize)
Write(*,*) '    cpos=',cpos, ' * nc=',ncluster

        If ((j==CacheSize) .AND. (k>Qmax)) then !*** Must break new cluster
Write(*,*) '   BREAKUP'
           s = OrgSize(i)

           k = Qmax 
           if (CacheSize > 0) k = k - ClusterCache(CacheSize)

           NCluster = Ncluster + 1
           ClusterMap(cpos+1:cpos+k) = NCluster
           cpos = cpos + k
           ClusterInfo(1,ncluster) = k
           If (CacheSize > 1) then
              ClusterInfo(2,ncluster) = SUM(ClusterCache(1:CacheSize-1))
           else
              ClusterInfo(2,ncluster) = 0
           End If
Write(*,*) '   nc=',ncluster, ' * add,drop=',ClusterInfo(:,ncluster), ' * left=',s-k
           CacheSize = 1
           ClusterCache(1) = Qmax
           s = s - k
           Do while (s > 0)
              k = Qmax - Qmin  !**ClusterCache(CacheSize)
              k = Min(k,s)
              NCluster = Ncluster + 1
              ClusterMap(cpos+1:cpos+k) = NCluster
              cpos = cpos + k

              ClusterInfo(1,ncluster) = k
              ClusterInfo(2,ncluster) = Min(k, Qmax-Qmin)

              s = s - k
Write(*,*) '   nc=',ncluster, ' * add,drop=',ClusterInfo(:,ncluster), ' * left=',s

           End do

           CacheSize = 2
           ClusterCache(1) = Qmax-Qmin
           ClusterCache(2) = Qmin

           Qsize = 0
        else     !*** can add the whole cluster ***
Write(*,*) '  WHOLE'
           Ncluster = NCluster+1
           ClusterMap(cpos+1:cpos+Orgsize(i)) = Ncluster
           Qsize = Orgsize(i)
           ClusterInfo(1,Ncluster) = Qsize
           ClusterInfo(2,ncluster) = SUM(ClusterCache(1:j-1))

           k = CacheSize-j+1
Write(*,*) '    k=',k, ' * j,cachesize=',j,cachesize
           ClusterCache(1:k) = ClusterCache(j:CacheSize)
           CacheSize = k+1
           ClusterCache(CacheSize) = OrgSize(i)
           cpos = cpos + OrgSize(i)
           Qsize = 0
        End if
Write(*,*) '     nc=',ncluster, ' * add,drop=',ClusterInfo(:,ncluster)
     else if (i==Orgclusters) then  !** Last cluster **
Write(*,*) '  LAST'
        Ncluster = NCluster+1
!        Qsize = Qsize + OrgSize(i)
        ClusterMap(cpos+1:cpos+Qsize) = Ncluster
!        ClusterInfo(1,Ncluster) = Qsize
        ClusterInfo(1,Ncluster) = OrgSize(i)
        ClusterInfo(2,Ncluster) = 0
     End if
  End Do

Write(*,*) 'JDQZ_Cluster: Final Size=',ncluster
j = 0
s=0; t=0;
Do i=1, Ncluster
   Write(*,*) 'JDQZ_Cluster: ',i, ' * add,drop=',ClusterInfo(:,i)
   
   Write(*,*) '  * OInit=',OrgMap(s+1:t)
   s = s + ClusterInfo(2,i)
   t = t + ClusterInfo(1,i)
   Write(*,*) '  *  OEnd=',OrgMap(s+1:t)
   j = j + ClusterInfo(1,i)
End do
Write(*,*) 'JDQZ_Cluster: Start ClusterMap=',OrgMap(1:nbands)
Write(*,*) 'JDQZ_Cluster: Final ClusterMap=',ClusterMap(1:nbands)

  return
End Subroutine



!******************************************************************************
! JDQZ_WindowBandSolver - Jacobi-Davidson Band solver
!  Uses B orthogonal vectors
!******************************************************************************

Subroutine JDQZ_WindowBandSolver(Emax, Target_shift, eigen, eivec, &
     ClusterInfo, EigIn, n, shift, eps, &
     kmax, jmax, jmin, Qmax, Qmin, method, m, L, maxnmv, maxstep, &
     maxperband, lock, order, work, lwork,out_unit,UseRandom)
   Use jd_misc
   Real,    Intent(IN)    :: Emax
   Real,    Intent(IN)    :: Target_Shift
   Real,    Intent(INOUT) :: Eigen(:)
   Complex, Intent(INOUT) :: eivec(:,:)
   Integer, Intent(IN)    :: ClusterInfo(:,:)
   Integer, Intent(IN)    :: EigIn
   Integer, Intent(IN)    :: N
   Complex, Intent(IN)    :: Shift
   Real,    Intent(IN)    :: eps
   Integer, Intent(INOUT) :: kmax
   Integer, Intent(IN)    :: jmax
   Integer, Intent(IN)    :: jmin
   Integer, Intent(IN)    :: Qmax
   Integer, Intent(IN)    :: Qmin
   Integer, Intent(IN)    :: method
   Integer, Intent(IN)    :: m
   Integer, Intent(IN)    :: L
   Integer, Intent(INOUT) :: maxnmv
   Integer, Intent(INOUT) :: maxstep
   Integer, Intent(INOUT) :: maxperband
   Real,    Intent(IN)    :: lock
   Integer, Intent(IN)    :: order
   Complex, Target, Intent(OUT)   :: Work(:,:)
   Integer, Intent(IN)    :: Lwork
   Integer, Intent(IN)    :: out_unit
   Logical, Intent(IN)    :: UseRandom

   Integer :: testspace, totalmv
   Complex :: alpha(jmax), beta(jmax)
   logical :: loop, try, found, ok
   integer :: ldvs, ldzwork, iseed(4),s,t, mode
   Complex :: ma(jmax,jmax), mb(jmax,jmax), zma(jmax,jmax), zmb(jmax,jmax)
   Complex :: vsl(jmax,jmax), vsr(jmax,jmax), ra(jmax,jmax), rb(jmax,jmax)
   Complex :: Gvsl(jmax,jmax), Gvsr(jmax,jmax)
!   Complex :: zwork(8*jmax), aconv(jmax), bconv(jmax)
   Complex :: zwork(4*jmax*jmax+7*jmax+1)
   integer :: ldqkz, copied, ncluster, clusterCopied, ClusterFound
   integer :: ipivqkz(jmax), NumToDrop, NumToKeep
   Complex :: mqkz(jmax,jmax), invqkz(jmax,jmax), f(jmax), dot

   integer :: i, j, k, info, mxmv, step, Qpos, Ktotal, bw, aq, zindex
   integer :: d, u, v, w, aux, av, bv, q, z, kz, itmp, tp, bq
   integer :: solvestep, LUT(jmax)
   Real    :: rnrm, rwork(8*Jmax), deps, dtmp, reig(jmax), Emin, tol
   Real    :: NewTarget, currerr, reig2(jmax)
   Complex :: zalpha, zbeta, targeta, targetb, evcond, shifta, shiftb
   Complex :: cdot, Wdot(Jmax)
   Complex :: dznrm2      !** External function 
   Complex, Pointer :: PWork(:)

   Integer :: CopyPos, NumFound, totalprec, totalgmresmv

   common /debug2/ idbg(20), outunit
   integer idbg, outunit
   integer :: kstr(4,4*jmax+24)
   Integer :: Index(8), struc(100), pstruc(4)
   Real    :: epsu, gap, adelta, bdelta
   Logical :: drop_zero, forceStore, FirstBand

   Complex, PARAMETER :: zero = (0,0)
   Complex, PARAMETER :: one = (1,0)
   data iseed /3,3,1966,29/            

   mode = 1
Write(*,*) 'JDQZ_WBS: Mode=',mode

   ldvs = jmax;  ldzwork = 8*jmax
   ldqkz = jmax
   ldzwork = 4*ldvs*ldvs + 7*ldvs + 1   + 4*ldvs

   !** Are there errors in the parameters?
!*   if ( kmax .gt. jmax ) call error('jdqz: kmax greater than jmax')
   if ( method .ne. 1 .and. method .ne. 2 ) &
        call error('jdqz: illegal choice for solution method')
   if ( order .lt. -2 .or. order .gt. 2 ) &
        call error('illegal value for order, must be between -2 and 2')

   d   = 1  !** d= rhs, these ptr's refer to the columns of the workspace
   tp  = d+1 !** Workspace for jdqzmv
   u   = tp+1 !** u   = pointer to Krylov space GMRES(m) or Bi-CSTAB(l)
   !**  v   = pointer to search space JDQZ with max dimension jmax
   if ( method .eq. JD_gmres ) then
      v   = u+m+1
   else if ( method .eq. JD_cgstab ) then
      v   = u+2*l+6
   end if
   w   = v+jmax !** pointer to test subspace JDQZ with max dimension jmax
   bw   = w+jmax !** pointer to test subspace JDQZ with max dimension jmax
   av  = bw+jmax !** pointer to subspace AV with max dimension jmax
   bv  = av+jmax  !** pointer to subspace BV with max dimension jmax
   aux = bv+jmax
   q   = aux+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
   aq  = q+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
   bq  = aq+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
!   wq  = bq+jmax
   z   = bq+kmax  !** pointer to test Schur basis in JDQZ with max dim kmax
   kz  = z+kmax  !** kz  = pointer to matrix K^{-1}Z_k

   if (kz+kmax-1.gt.lwork) then
      Write(out_unit,*) 'LWORK=',lwork, ' * should be ',kz+kmax-1
      call error ('qz: memory fault')
   End If

   !*** initialize loop **

   testspace = 3
   ok = .true.

   NewTarget = shift
   evcond = dcmplx(sqrt(abs(shift)**2+abs(one)**2))
   shifta = shift/evcond
   shiftb = one/evcond   
!   shifta = shift
!   shiftb = 1.0

   targeta = shifta
   targetb = shiftb
   Emin = -1E10

   zalpha = shifta
   zbeta = shiftb

   step = 0
   deps = dble(one)
   mxmv = 0

   solvestep = 0
   copied = 0
   NumFound = 0
   Qpos = 0
   CopyPos = 0
   Ktotal = 0

   j = 0
   k = 0
   totalmv = 0
   totalprec = 0
   totalgmresmv = 0
   ncluster = 1
   NumFound = Min(Jmin-1, ClusterInfo(1,ncluster))
   ClusterFound = 0
   ClusterCopied = 0

Write(*,*) 'Jmin=',Jmin, ' * Jmax=',Jmax, ' * Kmax=',Kmax, ' * Qmax=',Qmax, ' * Qmin=',Qmin
   !** Main Loop
   Loop = .TRUE.
   Do While (Loop)
      step = step+1
      solvestep = solvestep+1
      write(*,*) 'Outer step=',step, ' * SolveStep=',SolveStep, ' * j=',j, ' * copied=',copied, ' * EigIn=',EigIn, ' * K=',k, ' * Ktotal=',Ktotal, ' * NF=',NumFound
      Write(*,*) '     * a,b=',zalpha,zbeta
      Write(*,*) '     * eig=',(zalpha/zbeta)
      Write(*,*) '     sa,sb=',Shifta, shiftb, ' * sa/sb=',ShiftA/ShiftB
      Write(*,*) '     ta,tb=',Targeta, Targetb, ' * ta/tb=',Targeta/Targetb

      mxmv = 0
      if (j==0) then
         call zlarnv(2, iseed, n, work(1,v+j))
         call zlarnv(2, iseed, n, work(1,w+j))
         do i=1,n
            dtmp = dble(work(i,v+j))
            work(i,v+j) = dcmplx(dtmp,0d0)
         enddo
!         copied = copied + 1
!         Work(:,v+j) = EiVec(:,copied)
!         FirstBand = .TRUE.
         SolveStep = -MaxPerBand
!Q         NumFound = 1
      else 
         deps = 2d0**(-solvestep)
deps = eps
         if ((NumFound > 0) .AND. (copied<EigIn)) then
            copied = copied + 1
            Work(:,v+j) = EiVec(:,copied)           
            zalpha = Eigen(copied); zbeta = 1; 
            NumFound = NumFound - 1
            ClusterCopied = ClusterCopied + 1
Write(*,*) 'Adding CC=',ClusterCopied
            If (k == 0) then
               SolveStep = -MaxPerBand
            else
               SolveStep = 0
            End If
!Q            If (NumFound > 0) then
!Q               solvestep = 0
!Q            else
!Q               NumFound = 0
!Q            End if
              
!Q            SolveStep = 0
Write(*,*) 'JDQZ: AFTER FOUND j=',j, ' * copied=',copied, ' * SolveStep=',solvestep, ' * NumFound=',numfound, ' * Eigen=',REAL(Zalpha)
         else if (j.lt.jmin) then
!            If ((UseRandom) .OR. (copied >= EigIn)) then
Write(*,*) 'JDQZ:  ZGMRES INIT DDOT=',DOT_PRODUCT(Work(:,d), Work(:,d))
!               mxmv = 1
               SolveStep = 0
               mxmv = maxnmv
               call zgmres(n, work(1,v+j), work(1,d), m, deps, &
                    mxmv, zalpha, zbeta, k+1, &
                    work(1,kz), work(1,z), invqkz, ldqkz, &
                    ipivqkz, f, work(1,u), work(1,tp) )
!Q         else if ((solvestep == 0) .AND. (copied < EigIn))then
!Q            copied = copied + 1
!Q            Work(:,v+j) = EiVec(:,copied)           
!Q            NumFound = NumFound - 1
!Q            If (NumFound > 0) then
!Q               solvestep = 0
!Q            else
!Q               NumFound = 0
!Q           End if
              
!QWrite(*,*) 'JDQZ: AFTER FOUND j=',j, ' * copied=',copied, ' * SolveStep=',solvestep, ' * NumFound=',numfound           
         else if (method.eq.JD_gmres) then
Write(*,*) 'JDQZ:  ZGMRES DDOT=',DOT_PRODUCT(Work(:,d), Work(:,d))
Write(*,*) 'JDQZ:  ZGMRES ZDOT=',DOT_PRODUCT(Work(:,z+k), Work(:,z+k))
Write(*,*) 'JDQZ:  ZGMRES KZDOT=',DOT_PRODUCT(Work(:,kz+k), Work(:,kz+k))
            mxmv = maxnmv
            call zgmres(n, work(1,v+j), work(1,d), m, deps, &
                 mxmv, zalpha, zbeta, k+1, &
                 work(1,kz), work(1,z), invqkz, ldqkz, &
                 ipivqkz, f, work(1,u), work(1,tp) )
         else if (method.eq.JD_cgstab) then
Write(*,*) 'JDQZ:  ZCGSTABL'
            mxmv = maxnmv
            call zcgstabl(n, work(1,v+j), work(1,d), l, &
                 deps, mxmv, zalpha, &
                 zbeta, k+1, work(1,kz), work(1,z), invqkz, &
                 ldqkz, ipivqkz, f, work(1,u), 2*l+6)
         end if      
      end if
      
      TotalMV = TotalMV + mxmv
      TotalGMRESmv = TotalGMRESmv + mxmv
Write(*,*) 'step=',step, ' * TotalGMRESMV=',TotalGMRESMV

Write(*,*) '######## VDOT=',Dot_PRODUCT(Work(:,v+j),Work(:,v+j))
Write(*,*) '######## DDOT=',Dot_PRODUCT(Work(:,d),Work(:,d))

      j = j+1

      call hmgs (n, k, work(1,q), Work(1,bq), work(1,v+j-1))
Write(*,*) '######## After HMGS1 VDOT=',Dot_PRODUCT(Work(:,v+j-1),Work(:,v+j-1))      
!      call hmgs (n, j-1, work(1,v), work(1,bv), work(1,v+j-1))
      call hmgs (n, j-1, work(1,v), work(1,v), work(1,v+j-1)) !*QQ=I
Write(*,*) '######## After HMGS2 VDOT=',Dot_PRODUCT(Work(:,v+j-1),Work(:,v+j-1))      

if (j>1) write(*,*) 'DOT(Bv1)=',DOT_PRODUCT(Work(:,bv),Work(:,bv))
!k=0

Write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! j,k,copied=',j,k,copied
      !*** Normalize new V ***
      rnrm = REAL(DOT_PRODUCT(Work(:,v+j-1),Work(:,v+j-1)))  !*QQ=I
      rnrm = SQRT(rnrm)
      Work(:,v+j-1) = Work(:,v+j-1)/rnrm

      Call ABmul(n, work(1,v+j-1), work(1,av+j-1), work(1,bv+j-1))
      totalmv = totalmv + 1

      !*** Form new W and Normalize ***
Work(:,w+j-1) = Work(:,av+j-1) - eigen(copied)*Work(:,bv+j-1)
Write(*,*) 'Error=',DOT_PRODUCT(Work(:,w+j-1),Work(:,w+j-1))
      Work(:,w+j-1) = shiftb*Work(:,av+j-1) - shifta*Work(:,bv+j-1)
      call hmgs (n, k, work(1,q), Work(1,bq), work(1,w+j-1))      
      call zmgs (n, j-1, work(1,w), work(1,w+j-1), 1)

      call makemm (n, j, work(1,w), work(1,av), ma, zma, ldvs)
      call makemm (n, j, work(1,w), work(1,bv), mb, zmb, ldvs)
Write(*,*) 'after makemm'


Write(*,*) '@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'
      try = .true.
      if ((j<Jmin) .AND. .NOT. UseRandom) try = .FALSE.
      If ((.NOT. Try) .AND. (copied >= EigIn)) Try = .TRUE.

!Write(*,*) 'Ma----------------------------------------------'
!Do i=1, j
!   Write(*,*) 'JDQZ_WBS: i=',i, ' * ',Ma(i,1:j)
!End do
!Write(*,*) 'Mb----------------------------------------------'
!Do i=1, j
!   Write(*,*) 'JDQZ_WBS: i=',i, ' * ',Mb(i,1:j)
!End do

try=.true.   !** FIXME **
      If (try) then
         tol = 1E-10
!         Call FilterMatrix(j, ma, ldvs, tol)
!         Call FilterMatrix(j, mb, ldvs, tol)
         
         zma(1:j,1:j) = ma(1:j,1:j)
         zmb(1:j,1:j) = mb(1:j,1:j)
         epsu = 1E-10
         gap = 1000
         drop_zero = .FALSE.
         t = j
         Call GUPTRI(zMa(1,1), zMb(1,1), ldvs, t, t, epsu, gap, drop_zero, &
              GVSL(1,1), ldvs, GVSR(1,1), ldvs, &
              adelta, bdelta, index(1), index(2), index(3), index(4), &
              index(5), index(6), index(7), index(8), &
              pstruc(1), struc(1), zwork(1), kstr(1,1), info)
         
         if (info == 0) then
            Do i=1, j
               alpha(i) = zMa(i,i)
               beta(i) = zMb(i,i)
            End do
            reig(1:j) = alpha(1:j)/beta(1:j)
         End If

         i = index(8)-Index(2)
         If (i /= t) then
            Write(*,*) 'JDQZ_WBS: GUPTRI INCE-RTCE=',i, ' * Should be',t
            Write(*,*) 'JDQZ_WBS: Index=',Index
            STOP
         End If
         If (Info /= 0) then
            Write(*,*) 'JD_WindowBandSolver: GUPTRI =',info
            STOP
         End If

!Write(*,*) 'GUPTRI Eigen=',reig(1:j)
!Write(*,*) 'GUPTRI Matrix:  Ma'
!Do i=1, j
!   Write(*,*) 'i=',i, ' zma=',zma(i,1:j)
!End do
!Write(*,*) 'GUPTRI Matrix:  Mb'
!Do i=1, j
!   Write(*,*) 'i=',i, ' zma=',zmb(i,1:j)
!End do

         Call ZGEGV('V', 'V', j, zMa(1,1), LDVS, zMb(1,1), LDVS,  &
              alpha, beta, Ra, ldvs, Rb, ldvs, zwork, ldzWork, Rwork, info)
         if (info == 0) reig(1:j) = alpha(1:j)/beta(1:j)

         If (Info /= 0) then
            Write(*,*) 'JD_WindowBandSolver: ZGEGV =',info
            STOP
         End If

         VSL(1:j,1:j) = MATMUL(GVSL(1:j,1:j), Ra(1:j,1:j))
         VSR(1:j,1:j) = MATMUL(GVSR(1:j,1:j), Rb(1:j,1:j))
Write(*,*) 'Raw eig=', reig(1:j), ' * TARGET=',REAL(targeta/targetb)

!         zma(1:j,1:j) = ma(1:j,1:j)
!         zmb(1:j,1:j) = mb(1:j,1:j)
!         Call ZGEGV('V', 'V', j, zMa(1,1), LDVS, zMb(1,1), LDVS,  &
!              alpha, beta, VSL, ldvs, VSR, ldvs, zwork, ldzWork, Rwork, info)
!        if (info == 0) reig(1:j) = alpha(1:j)/beta(1:j)
!        if (info /= 0) then
!           Write(*,*) 'ZGEGV Test Error=',info
!           STOP
!        End If
!Write(*,*) 'ZGEGV eig=', reig(1:j), ' * TARGET=',REAL(targeta/targetb)
      End If


!Ra(1:j,1:j) = MATMUL(Ma(1:j,1:j), VSR(1:j,1:j))
!Rb(1:j,1:j) = MATMUL(Mb(1:j,1:j), VSR(1:j,1:j))
!Do i=1, j
!   work(1:j, u) = Ra(1:j,i) - reig(i)*Rb(1:j,i)
!   Write(*,*) 'ZGEGV i=',i, ' * Hx-eOx=',DOT_PRODUCT(Work(1:j,u), Work(1:j,u)) 
!   Do t=1, j
!      Write(*,*) ' ij=',i,t, ' * <i|O|j>=',DOT_PRODUCT(VSR(1:j,i), Rb(1:j,t))
!   End Do
!end Do
      Do While (try)
!k=0 !FIXME
         !*** Sort the Petrov pairs ***
         Where (reig(1:j) < Emin) 
            reig(1:j) = 1E10
            alpha(1:j) = 1E10; beta(1:j) = 1
         End Where
!**         alpha(1:j) = reig(1:j); beta = 1

zwork(1:j) = alpha(1:j);  zwork(j+1:2*j) = beta(1:j)
         If (JDQZ_Spatial_Selection) then
            alpha(1:j) = 0;  beta(1:j) = 1.0
            Pwork => SCFvalues%Work
            Do i=1, j
               call zgemv ('n', n, j, one, work(1,v), n, vsr(1,i), &
                    1, zero, work(1,aux), 1)

               Pwork = 0               
               Call RS_toR(Work(:,aux), Pwork)
               alpha(i) = SUM(CONJG(Pwork)*Local_Criteria_Map*PWork)
               alpha(i) = alpha(i)/SUM(CONJG(Pwork)*Pwork)
            End Do
Write(*,*) 'JDQZ: Spatial Wts=',REAL(alpha(1:j)), ' * Order=',order
         End If
!alpha(1:j) = zwork(1:j); beta(1:j) = zwork(j+1:2*j)   !**FIXME

         Call JD_Sort(REAL(targeta/targetb), j, alpha, beta, LUT, Order)
!*         alpha(1:j) = zwork(1:j); beta(1:j) = zwork(j+1:2*j)   !**FIXME

         Ra = VSL; Rb = vsr
!**         zwork(1:j) = alpha(1:j);  zwork(j+1:2*j) = beta(1:j)
!**         zwork(1:j) = Reig(1:j);  zwork(j+1:2*j) = 1.0
         Do i=1, j
            alpha(i) = zwork(LUT(i))
            beta(i) = zwork(j+LUT(i))
            vsl(:,i) = Ra(:,LUT(i))
            vsr(:,i) = Rb(:,LUT(i))
         End do
         reig(1:j) = alpha(1:j)/beta(1:j)
         
Write(*,*) 'After QZSORT!!!!! eig=', reig(1:j), ' * TARGET=',REAL(targeta/targetb), ' * LUT=',LUT(1:j)

         zalpha = alpha(1);  zbeta = beta(1);
         
         evcond = dcmplx(sqrt(abs(zalpha)**2+abs(zbeta)**2))
         
         !*** compute new q ***
         call zgemv ('n', n, j, one, work(1,v), n, vsr(1,1), &
              1, zero, work(1,q+k), 1)
         call hmgs (n, k, work(1,q), work(1,bq), work(1,q+k))

write(*,*) 'Q k+1=',k+1, ' DOT=',DOT_PRODUCT(Work(:,q+k),Work(:,q+k))
         !*** Calc new Aq and Bq(z) ***
         Call ABMul(n, work(1,q+k), Work(1,aq+k), Work(1,bq+k))
         rnrm = DOT_PRODUCT(Work(:,q+k), Work(:,bq+k))
         rnrm = sqrt(ABS(rnrm))
         Work(:,q+k) = Work(:,q+k)/rnrm
         Work(:,aq+k) = Work(:,aq+k)/rnrm
         Work(:,bq+k) = Work(:,bq+k)/rnrm

cdot = DOT_PRODUCT(Work(:,q+k), Work(:,aq+k))
write(*,*) 'Bq k+1=',k+1, ' DOT=',DOT_PRODUCT(Work(:,bq+k),Work(:,bq+k))
         
         !*** Make new qkz ***
         If (mode == 1) then
            Work(:,z+k) = Work(:,bq+k) !QW1
         else if (mode == 2) then
            Work(:,z+k) = Work(:,q+k)  !QW2
         End If
         Work(:,kz+k) = Work(:,z+k)
         call precon(n, work(1,kz+k))
         totalprec = totalprec + 1
         call mkqkz (n, k+1, work(1,z), work(1,kz), mqkz, invqkz, &
              ldqkz, ipivqkz)

         !*** compute new (right) residual= beta Aq - alpha Bq and
         !*** orthogonalize this vector on Z.
         Work(:,d) = zbeta*Work(:,aq+k) - zalpha*Work(:,bq+k)
         rnrm = dznrm2 (n, work(1,d), 1)/dble(evcond)
         call hmgs (n, k, work(1,q), Work(1,bq), work(1,d))

currerr = rnrm
Write(*,*) 'CURRERR=',rnrm
Write(*,*) 'DDOT=',Dot_PRODUCT(Work(:,d),Work(:,d))
         if (rnrm.lt.lock.and.ok) then
!            targeta = zalpha
!            targetb = zbeta
            ok = .false.
         end if
         
         found = (rnrm.lt.eps.and.(j.gt.1.or.k.eq.kmax-1))
!**ADDED If last step then force return of rest of eigenvects
         if (step .EQ. MaxStep) then
            found = (ktotal.le.kmax-1) 
         End If
!**ADDED
if ((SolveStep >= MaxPerBand) .AND. .NOT. Found) then
   found = (ktotal.le.kmax-1) 
   write(*,*) 'JDQZ: Forcing STORE!'
End If

         if (SolveStep >= MaxPerBand) found = (ktotal<=kmax-1)

         try =  found
         
         if (found) then              
            FirstBand = .FALSE.
            !*** increase the number of found evs by 1 ***
            k = k+1
            Ktotal = Ktotal + 1
            Qpos = Qpos + 1
            ClusterFound = ClusterFound + 1
!*            ClusterCopied = ClusterCopied + 1
Write(*,*) 'FOUND START CF=',ClusterFound, ' * CC=',ClusterCopied
Write(*,*) '   NC=',Ncluster, ' * CI=',ClusterInfo(:,ncluster), ' * NF=',NumFound
            If (Clusterfound == ClusterInfo(1,ncluster)) then
               ncluster = ncluster + 1
               NumFound = Min(Jmin-1, ClusterInfo(1,ncluster))
               NumToDrop = ClusterInfo(2,ncluster)
               NumToKeep = k - NumToDrop
               ClusterFound = 0
               ClusterCopied = 0
            else
               NumToDrop = 0
               NumToKeep = k
               if (ClusterCopied < ClusterInfo(1,ncluster)) then
                  NumFound = NumFound + 1
               End if
            End if
Write(*,*) 'FOUND END CF=',ClusterFound, ' * CC=',ClusterCopied
Write(*,*) '   NC=',Ncluster, ' * CI=',ClusterInfo(:,ncluster)
Write(*,*) '   NTD=',NumtoDrop, ' * NTK=',NumToKeep, ' * NF=',NumFound

write(*,*) 'FOUND step=',step, ' * k=',k, ' * eig=',REAL(zalpha/zbeta), ' * CalcEig=',REAL(cdot), ' * ERR=',currerr, ' * CF=',ClusterFound, ' *nc=',ncluster, ' * CI(:,nc)=',Clusterinfo(:,ncluster)
            
            !*** store the eigenvalue
            Eigen(ktotal) = zalpha/zbeta
!            Eigen(ktotal) = DOT_PRODUCT(Work(q+k,:), Work(aq+k,:))

            solvestep = 0
            If (copied < EigIn) solvestep = -1

            If (ktotal < Kmax) then
               If (NumToDrop > 0) then               
!Write(*,*) 'DROP Start Eigens=',Eigen(1:Ktotal)
!                 !** Do a subspace rotation of these vectors before storing **
!                  Do i=1, k
!                     Do t=1, i
!                        Ra(i,t) = DOT_PRODUCT(Work(:,q+i-1), Work(:,aq+t-1))
!                        Rb(i,t) = DOT_PRODUCT(Work(:,q+i-1), Work(:,bq+t-1))
!                        If (i /= t) then
!                           Ra(t,i) = CONJG(Ra(i,t))
!                           Rb(t,i) = CONJG(Rb(i,t))
!                        End If
!                     End do
!                  End do
!                  Call ZHEGV(1, 'V', 'L', k, Ra(1,1), LDVS, Rb(1,1), &
!                       LDVS, reig2, zwork, ldzWork, Rwork, info)
!
!                  If (Info /= 0) then
!                     Write(*,*) 'JDQZ_WindowBandSolver: ZHEGV =',info
!                     STOP
!                  End If
!
!                  Eigen(Ktotal-k+1:Ktotal) = reig2(1:k)
!                  Work(:,aux:aux+k-1) = MATMUL(Work(:,q:q+k-1), Ra(1:k,1:k))
!                  t = q; q = aux; aux = t;
!                  Work(:,aux:aux+k-1) = MATMUL(Work(:,aq:aq+k-1), Ra(1:k,1:k))
!                  t = aq; aq = aux; aux = t;
!                  Work(:,aux:aux+k-1) = MATMUL(Work(:,bq:bq+k-1), Ra(1:k,1:k))
!                  t = bq; bq = aux; aux = t;
!Write(*,*) 'DROP After Eigens=',Eigen(1:Ktotal)
!Write(*,*) 'DROP ZHEGV Eigens=',reig2(1:K)
                  
                  NewTarget = Reig(1) + Target_shift
                  If (copied < EigIn) then
                     Newtarget = 0.05*(eigen(Copied+1)+Target_Shift) + &
                                 0.95*Newtarget
                  End If

                  Emin = 0.8*MinVal(Eigen(Ktotal-NumtoKeep+1:Ktotal)) + &
                         0.2*MaxVal(Eigen(Ktotal-NumToKeep+1:Ktotal))
Write(*,*) 'DROP: Dropping=',Numtodrop, ' * Keeping=',NumTokeep
Write(*,*) 'DROP: EigenSearch=',Eigen(Ktotal-NumToKeep+1:Ktotal), ' * Min=', &
    MinVal(Eigen(Ktotal-NumToKeep+1:Ktotal)), &
    'Max=',MaxVal(Eigen(Ktotal-NumToKeep+1:Ktotal)), &
    ' Emin(AV)=',Emin, ' *  NewTarget=',NewTarget, ' * Eig(+1)=',eigen(copied+1)
                  Emin = Min(Emin, NewTarget)
                
                  evcond = dcmplx(sqrt(abs(NewTarget)**2+abs(one)**2))
                  shifta = NewTarget/evcond
                  shiftb = one/evcond   
!                  shifta = NewTarget
!                  shiftb = 1.0
                  targeta = shifta; targetb = shiftb                  
               End If

               call zgemm ('n', 'n', n, j-1, j, one, work(1,v), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = v
               v = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,w), n, &
                    vsl(1,2), ldvs, zero, work(1,aux), n)
               itmp = w
               w = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,bw), n, &
                    vsl(1,2), ldvs, zero, work(1,aux), n)
               itmp = bw
               bw = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,av), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = av
               av = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,bv), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = bv
               bv = aux
               aux = itmp
               j = j-1
               zma = 0;  ma=0;  mb=0;  zmb=0; vsr=0; vsl=0


               Do i=0, j-1
Work(:,u) = Work(:,v+i)
                  Call mgs_weight(n, k, work(1,q), Work(1,bq), &
                       work(1,v+i),wdot)
                  Work(:,av+i) = Work(:,av+i) - &
                       MATMUL(Work(:,aq:aq+k-1), Wdot(1:k))
                  Work(:,bv+i) = Work(:,bv+i) - &
                       MATMUL(Work(:,bq:bq+k-1), Wdot(1:k))
!*                  call hmgs(n, k, work(1,q), Work(1,bq), work(1,v+i))      
Work(:,u) = Work(:,u) - Work(:,v+i)
Write(*,*) 'i=',i,' * change in V=',DOT_PRODUCT(Work(:,u), Work(:,u))
!call hmgs (n, i, work(1,v), Work(1,bv), work(1,v+i))      
                  call mgs_weight(n, i, work(1,v), Work(1,v), &
                       work(1,v+i), wdot) !*QQ=I
                  If (i>0) then
                     Work(:,av+i) = Work(:,av+i) - &
                          MATMUL(Work(:,av:av+i-1), Wdot(1:i))
                     Work(:,bv+i) = Work(:,bv+i) - &
                          MATMUL(Work(:,bv:bv+i-1), Wdot(1:i))
                  end If

!                  call hmgs (n, i, work(1,v), Work(1,v), work(1,v+i)) !*QQ=I
!Call ABmul(n, work(1,v+i), work(1,av+i), work(1,bv+i))

                  rnrm = DOT_PRODUCT(Work(:,v+i), Work(:,bv+i))
rnrm = DOT_PRODUCT(Work(:,v+i), Work(:,v+i)) !*QQ=I
                  rnrm = SQRT(rnrm)
!Write(*,*) 'ReNorm: VMAG=',rnrm
                  Work(:,v+i) = Work(:,v+i)/rnrm
                  Work(:,av+i) = Work(:,av+i)/rnrm
                  Work(:,bv+i) = Work(:,bv+i)/rnrm
Work(:,bw+i) = Work(:,w+i) !*QQ=I
                  rnrm = DOT_PRODUCT(Work(:,w+i), Work(:,bw+i))
                  rnrm = SQRT(rnrm)
                  Work(:,w+i) = Work(:,w+i)/rnrm
                  Work(:,bw+i) = Work(:,bw+i)/rnrm

!Write(*,*) 'ReNorm: WMAG=',rnrm
                  Work(:,w+i) = ShiftB*Work(:,av+i) - Shifta*Work(:,bv+i)
!Work(:,w+i) = Work(:,v+i)
!**                  call zmgs (n, k, work(1,q), work(1,w+i), 1)
                  call hmgs (n, k, work(1,q), Work(1,bq), work(1,w+i))      
                  call zmgs (n, i, work(1,w), work(1,w+i), 1)

!Do t=0,i
!   Write(*,*) 'i,j=',i,t, ' * <i|O|j>=',DOT_PRODUCT(Work(:,w+i), Work(:,bw+t))
!End Do
               End Do

               Do i=1, j
Write(*,*) 'i=',i, ' * NEW vAv/vBv=',REAL(DOT_PRODUCT(Work(:,v+i-1), Work(:,av+i-1))/DOT_PRODUCT(Work(:,v+i-1), Work(:,bv+i-1)))
                  Do t=1, j
                     dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,av+t-1))
                     zma(i,t) = dot
                     ma(i,t) = dot
                     dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,bv+t-1))
                     zmb(i,t) = dot
                     mb(i,t) = dot
                  End do
!Write(*,*) 'JDQZ_WBS: i=',i, ' * Ma(i,:)=',Ma(i,1:j)
!Write(*,*) 'JDQZ_WBS: i=',i, ' * Mb(i,:)=',Mb(i,1:j)
                  vsr(i,i) = 1
                  vsl(i,i) = 1
                  alpha(i) = alpha(i+1)
                  beta(i) = beta(i+1)
               End do

zma(1:j,1:j) = ma(1:j,1:j)
zmb(1:j,1:j) = mb(1:j,1:j)
epsu = 1E-10
gap = 1000
drop_zero = .FALSE.
t = j
Call GUPTRI(zMa(1,1), zMb(1,1), ldvs, t, t, epsu, gap, drop_zero, &
     GVSL(1,1), ldvs, GVSR(1,1), ldvs, &
     adelta, bdelta, index(1), index(2), index(3), index(4), &
     index(5), index(6), index(7), index(8), &
     pstruc(1), struc(1), zwork(1), kstr(1,1), info)
Write(*,*) 'After reorg GUPTRI Eigen'
Do i=1,j
   Write(*,*) ' i=',i, ' * Eig=',zma(i,i)/zmb(i,i)
End do

!               reig(1:j) = alpha(1:j)
!               targeta = shifta
!               targetb = shiftb
               ok = .true.
               mxmv = 0
               deps = dble(one)

               !*** Check for dropping ***
               If (NumToDrop > 0) then
Write(*,*) 'DROPPING:  k>Qmax: k=',k

                  Do i=1, (k-NumToKeep)
                     Copypos = CopyPos + 1
                     EiVec(:,CopyPos) = Work(:,q+i-1)
                  End do
                  i = k-NumToKeep
                  Work(:,q:q+NumToKeep-1) = Work(:,q+i:q+k-1)
                  Work(:,bq:bq+NumToKeep-1) = Work(:,bq+i:bq+k-1)
                  k = NumToKeep

Write(*,*) 'DROPPING:  CopyPos=',CopyPos, ' * New Energy=',NewTarget, ' * Emin=',Emin
                  Call MakePrecondMatrix(NewTarget)

                  
Write(*,*) 'DROPPING: Eigens Kept=',Eigen(CopyPos+1:Ktotal)
Write(*,*) 'DROPPING: Eigens ALL =',Eigen(1:Ktotal)
                  !** Now reform the zKz Matrix ***
                  mqkz(1:NumToKeep,1:NumToKeep) = 0
                  invqkz(1:NumToKeep,1:NumToKeep) = 0
                  Do i=1, NumToKeep
Call ABMUL(n, Work(1,q+i-1), Work(1,u), Work(1,tp))
Write(*,*) 'DROPPING: i=',i, ' * CalcEig=',REAL(DOT_PRODUCT(Work(:,q+i-1), Work(:,u))), ' * Mag=',REAL(DOT_PRODUCT(Work(:,q+i-1), Work(:,tp)))
                     If (mode == 1) then
                        Work(:,z+i-1) = Work(:,bq+i-1)  !QW1
                     else if (mode == 2) then
                        Work(:,z+i-1) = Work(:,q+i-1) !QW2
                     end If
                     Work(:,kz+i-1) = Work(:,z+i-1)
                     call precon(n, work(1,kz+i-1))
                     totalprec = totalprec + 1
                    
                     call mkqkz (n, i, work(1,z), work(1,kz), mqkz, &
                          invqkz, ldqkz, ipivqkz) 
                  End do

!Do i=1, j
!  Call ABMUL(n, Work(1,v+i-1), Work(1,u), Work(1,tp))
!Write(*,*) 'DROPPING: VLEFT i=',i, ' * CalcEig=',REAL(DOT_PRODUCT(Work(:,v+i-1), Work(:,u))), ' * Mag=',REAL(DOT_PRODUCT(Work(:,v+i-1), Work(:,tp)))
!end Do
               End If
!j=0
!Write(*,*) 'RESETTING J!!!!!!'
!try = .FALSE.
            else
               try = .FALSE.
            end If
         else if (j.eq.jmax) then
Write(*,*) 'SHRINKING!!!!!!!!!!!!!!!!!!!!!!!!!'
            call zgemm ('n', 'n', n, jmin, j, one, work(1,v), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = v
            v = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,av), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = av
            av = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,bv), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = bv
            bv = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,w), n, &
                 vsl, ldvs, zero, work(1,aux), n)
            itmp = w
            w = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,bw), n, &
                 vsl, ldvs, zero, work(1,aux), n)
            itmp = bw
            bw = aux
            aux = itmp

!--------------
            j = jmin
            zma = 0;  ma=0;  mb=0;  zmb=0; vsr=0; vsl = 0
            Do i=0, j-1
               Call mgs_weight(n, k, work(1,q), Work(1,bq), &
                    work(1,v+i),wdot)
               Work(:,av+i) = Work(:,av+i) - &
                    MATMUL(Work(:,aq:aq+k-1), Wdot(1:k))
               Work(:,bv+i) = Work(:,bv+i) - &
                    MATMUL(Work(:,bq:bq+k-1), Wdot(1:k))
               
               call mgs_weight(n, i, work(1,v), Work(1,v), &
                    work(1,v+i), wdot) !*QQ=I
               If (i>0) then
                  Work(:,av+i) = Work(:,av+i) - &
                       MATMUL(Work(:,av:av+i-1), Wdot(1:i))
                  Work(:,bv+i) = Work(:,bv+i) - &
                       MATMUL(Work(:,bv:bv+i-1), Wdot(1:i))
               end If


               rnrm = DOT_PRODUCT(Work(:,v+i), Work(:,v+i)) !*QQ=I
               rnrm = SQRT(rnrm)
               Work(:,v+i) = Work(:,v+i)/rnrm
               Work(:,av+i) = Work(:,av+i)/rnrm
               Work(:,bv+i) = Work(:,bv+i)/rnrm

               Work(:,w+i) = ShiftB*Work(:,av+i) - Shifta*Work(:,bv+i)
               call hmgs (n, k, work(1,q), Work(1,bq), work(1,w+i))      
               call hmgs (n, i, work(1,w), Work(1,w), work(1,w+i))

               rnrm = REAL(DOT_PRODUCT(Work(:,w+i),Work(:,w+i)))
               rnrm = SQRT(rnrm)
               Work(:,w+i) = Work(:,w+i)/rnrm
            End Do

            Do i=1, j
               Do t=1, j
                  dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,av+t-1))
                  zma(i,t) = dot
                  ma(i,t) = dot
                  dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,bv+t-1))
                  zmb(i,t) = dot
                  mb(i,t) = dot
               End do

               vsr(i,i) = 1
               vsl(i,i) = 1
            End do
!--------------

!            j = jmin
!            zma = 0;  ma=0;  mb=0;  zmb=0; vsr=0;
!            Do i=1, j
!               Do t=1, j
!                  dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,av+t-1))
!                  zma(i,t) = dot
!                  ma(i,t) = dot
!                  dot = DOT_PRODUCT(Work(:,w+i-1), Work(:,bv+t-1))
!                  zmb(i,t) = dot
!                  mb(i,t) = dot
!               End do
!!Write(*,*) 'JDQZ_WBS: i=',i, ' * Ma(i,:)=',Ma(i,1:j)
!!Write(*,*) 'JDQZ_WBS: i=',i, ' * Mb(i,:)=',Mb(i,1:j)
!            
!               vsr(i,i) = 1
!               vsl(i,i) = 1
!!               alpha(i) = alpha(i+1)
!               beta(i) = beta(i+1)
!            End do

!            Do i=1, j
!               zma(i,i) = reig(i)
!               ma(i,i) = reig(i)
!               zmb(i,i) = 1
!               mb(i,i) = 1
!               vsr(i,i) = 1
!            End do
         end if
      End Do   !** TRY **
      
      loop = (ktotal.lt.kmax.and.step.lt.maxstep)
   End Do  !** End Big Loop ***
   

   !** Did enough eigenpairs converge?
   kmax = ktotal


   EiVec(:,copypos+1:copypos+k) = Work(:,q:q+k-1)

   Pwork => SCFvalues%Work
   Do i=1, Ktotal
      Pwork = 0               
      Call RS_toR(Eivec(:,i), Pwork)
      dtmp = SUM(CONJG(Pwork)*PWork)
      Pwork = CONJG(Pwork)*Local_Criteria_Map*PWork
      rnrm = SUM(PWork)
      dtmp = rnrm/dtmp
      Write(*,*) 'JDQZ: Final i=' ,i,' Frac=',dtmp, ' * eig=',Eigen(i)
   End Do


   maxnmv = totalmv
   Write(*,*) 'JDQZ_BS: FOUND TotalMV = ',TotalMV, ' * TotalPREC = ', TotalPrec
   Write(*,*) 'JDQZ_BS: FOUND TotalGMRESMV = ',TotalGMRESMV
   Write(*,*) 'JDQZ_BS: FOUND Eigen=',Eigen(1:kTotal)
 
   Return
End Subroutine




End Module




