!******************************************************************************
!
! File : solver.f90
!   by : Alan Tackett
!   on : 05/06/97
!  for : PAW Project
!
!  Module implementing the  Poisson/Eigenvalue problem that arises
!  from the PAW mothod.
!
!
!******************************************************************************

Module solver

Use anderson_mixing
Use atom_data
Use paw_inout
Use gpoints
Use mem_data
Use memmgr
Use psilib
Use fileio
Use search_sort
Use hamiltonian
Use debug
Use grrp
Use relax
Use storedata
Use lcao
Use timing
Use options_data

Implicit NONE!!!!!!!!!!!!!

Real, PRIVATE :: Alpha
Integer,      PRIVATE :: BG_Mode

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
! UpdateKpnt - Updates the Wave functions for a given Kpnt
!
!  Kpnt  - Kpnt index
!  VecSize - Number of Psi's for the kpoint
!  Basis_toProcess - List of Psi's to process
!
!******************************************************************************

Subroutine UpdateKpnt( Kpnt, VecSize, Basis_toProcess)
  Integer, Intent(IN) :: Kpnt
  Integer, Intent(IN) :: VecSize
  Integer, Intent(IN) :: Basis_toProcess(:)

  Complex  :: Vec(VecSize, VecSize)
  Real     :: Eigen(VecSize)
  Integer          :: LUT(2,Mem_MapSize), KpntBands, i, NewSize

  KpntBands = 0
  Do i=1, Mem_MapSize
     If ((PsiInfo(i)%Kpnt == Kpnt) .AND. (CLusterMap(i) > 0)) then
        KpntBands = KpntBands + 1
     End If
  End Do

  Call PerformGRRP_ALL( Basis_toProcess, KpntBands, &
       Vec, Eigen, LUT, NewSize, Kpnt)
  Call MakeNewPsi( KpntBands, NewSize, Vec, Eigen, LUT, Kpnt)

write(Log_Unit,*) 'completed updatekpnt'
  Return
End Subroutine



!******************************************************************************
!
!  CalcClusterRange - Calculates the cluster energy ranges
!
!    Energy - Energy ranges for each cluster
!
!******************************************************************************

Subroutine CalcClusterRange(Ecluster)
  Real, Intent(OUT) :: Ecluster(:,:)

  Integer :: i,j, OldKpnt, Kpnt, ClusterKpnt(LastCLuster+1), KpntStart, KpntEnd
  Real :: Epsi, Energy(2,LastCluster), Elo, Ehi

  ClusterKpnt = -1
  Energy(1,:) = 1E30
  Energy(2,:) = -1E30

  !** First determine the Energy ranges for each cluster ***
  Do i=1, Mem_MapSize
     If (ClusterMap(i) > 0) then
        j = ClusterMap(i)
        Epsi = PSiInfo(i)%Energy
        ClusterKpnt(j) = PsiInfo(i)%Kpnt
Write(Log_Unit,*) 'CalcClusterRange: i=',i, ' * j=',j, ' * E=',Epsi

        If (Epsi < Energy(1,j)) Energy(1,j) = Epsi
        If (Epsi > Energy(2,j)) Energy(2,j) = Epsi
     End If 
  End Do

Write(Log_Unit,*) 'CalcClusterRange: LastCluster = ',LastCluster
Do i=1, LastCluster
  Write(Log_Unit,*) 'CalcClusterRange: i=',i, ' * L,H=',Energy(:,i)
End Do


!Write(Log_Unit,*) 'ClusterKpnt=',ClusterKpnt
  KpntStart = 1
  Do Kpnt = 1, NumKpnts
     j = KpntStart
     Do While (ClusterKpnt(j) == Kpnt)
        j = j + 1
     End Do

     KpntEnd = j - 1

     If ( KpntStart < LastCluster) then
        Ecluster(1,KpntStart) = 0.5*(Energy(2,KpntStart)+Energy(1,KpntStart+1))
     else
        Ecluster(1,KpntStart) = 1E30
     End if

     If (KpntStart /= KpntEnd) then
        Ecluster(1,Kpntend) = Energy(2,KpntEnd)
        Elo = 0.5*(Energy(2,KpntEnd-1)+Energy(1,Kpntend))
        Ecluster(2,KpntEnd) = Energy(2,Kpntend) - Elo
     End If

     Do i=KpntStart+1, KpntEnd-1
        Elo = 0.5*(Energy(2,i-1) + Energy(1,i))
        Ehi = 0.5*(Energy(2,i) + Energy(1,i+1))
        Ecluster(1,i) = 0.5*(Elo + Ehi)
        Ecluster(2,i) = Ehi - Ecluster(1,i)
     End Do

     KpntStart = KpntEnd + 1
  End Do

  Do i=1, LastCluster
     Write(Log_Unit,*) 'CalcClusterRange: Cluster=',i, ' * Kpnt=', &
          CLusterKpnt(i)
     Write(Log_Unit,*) 'CalcClusterRange:  Eo=',Ecluster(1,i), &
          ' * Ew=',Ecluster(2,i)
  End Do
  Return
End Subroutine

!******************************************************************************
!
!  Cluster_Eigenvalues - Clusters the eigenvalues given and determines the
!         cut-off point to keep
!
!  MinPsi     - Min. number of PSi's to keep
!  NumPsi     - Acutal Number kept
!  Eigen      - Eigenvalue list
!  EigenSize  - Size of Eigenvalue list
!
!******************************************************************************

Subroutine Cluster_EigenValues(MinPsi, NumPsi, Eigen, EigenSize)
  Integer,      Intent(IN)  :: MinPsi
  Integer,      Intent(OUT) :: NumPsi
  Real,         Intent(IN)  :: Eigen(:)
  Integer,      Intent(IN)  :: EigenSize
  
  Integer :: i,j,k
  


!5-01-01-changed !** Only want 1 Psi in cluster beyond MINBANDS wave functions
!5-01-01-changed  i = MinPsi+1
   i=1
  Do While ((i<EigenSize) .AND. (i<TotalPsi) .AND. &
       (Eigen(i)  < Eigen_Max))
     i = i + 1
  End Do
  Do While ((i<EigenSize) .AND. (i<TotalPsi) .AND. &
       (ABS(Eigen(i) - Eigen(i+1)) < ClusterTol))
     i = i + 1
  End Do

Write(Log_Unit,*) 'Cluster_EigenValues: MinPsi=',MinPsi, ' * i=',i, ' * TotalPsi=',TotalPsi
  If (i>EigenSize) i = EigenSize
  If (i>TotalPsi) i = TotalPsi

  NumPsi = i

  Return
End Subroutine



!*****************************************************************************
!
! CleanUp_MemMap
!
!  Free unused virtual and buffer memory slots
!
!*****************************************************************************
Subroutine CleanUp_MemMap
Integer :: i,j,k

 Do i=1,Mem_MapSize
   If (PsiInfo(i)%Kpnt < 0) then
      PsiInfo(i)%Kpnt=0
      PsiInfo(i)%DoSave=0
      PsiInfo(i)%Available=Mem_Available
      PsiInfo(i)%PDOT_Stored=.FALSE.    
      PsiInfo(i)%KE_Stored=.FALSE.    
      PsiInfo(i)%OxIndex=0
      PsiInfo(i)%CorrIndex=0
      j=PsiInfo(i)%MemBufIndex
      If (j>0) Globalmap%Psi_Handle(j)%Index=0   
      PsiInfo(i)%MemBufIndex=0
   EndIf
 Enddo
Return

End Subroutine



!******************************************************************************
!
!  BlockDavidson  - Performs an update of V and Psi one cluster at a time
!
!  MaxIter   - Max number of iterations
!  EnergyTol - Energy tolerance required before completion
!  Converged - DEtemines if converged or not(RETURNED)
!  Iter      - Number of iterations
!
!******************************************************************************

Subroutine BlockDavidson( MaxIter, EnergyTol, Converged, iter)
  Integer,          Intent(IN)    :: MaxIter
  Real,             Intent(INOUT) :: EnergyTol
  Logical,          Intent(OUT)   :: Converged
  Integer,          Intent(OUT)   :: Iter

  Integer :: i,j,k, n, c, CorrType
  Integer :: VecProcessed(Mem_MapSize),  Vec_toProcess(Mem_MapSize)
  Integer :: ArraySize,  Kpnt, Psi_list(Mem_MapSize)
  Real :: EnergyDif, OldEnergy, StartPotError, PotError, E1, E2, E3
  Real :: NewEigen_Map(Mem_MapSize)
  Complex :: term
  Complex, Pointer :: Hmat(:,:), Omat(:,:), Vec(:,:)
  Real, Pointer :: Eigen(:)
  Complex, Pointer :: Hx(:), Ox(:), PsiWork(:), Ox2(:)
  Integer :: LUT(2, Mem_MapSize), LUT_Count, PsiIndex, Kindex, VecSize
  Integer :: Cluster_Start, Cluster_End, Cluster,NewSize
  Complex :: PDot(PLM_Max)
  Type(Mem_Handle), Pointer :: Psi2
  Character*100 :: ext
  Logical :: DoOrthog

Complex, Pointer :: B(:), X(:), Work(:)
Real :: tmp
Logical :: PrintUpdate

PrintUpdate = .FALSE.

PsiInfo(:)%Error = -1

  DoOrthog = SCF_DogramSchmidt

Write(Log_Unit,*) 'BlockDavidson : Start'
Call Print_WorkStatus(Log_Unit)

  ArraySize = -1
  Nullify(Hmat);  Nullify(Omat);  Nullify(Vec); Nullify(Eigen);

  BG_Mode = 0
  EnergyDif = 2*EnergyTol
  iter = 0
  Do While ((EnergyDif > EnergyTol) .AND. (iter < MaxIter))
     Call Start_Timer(Timer(Iter_Timer))
     iter = iter + 1


     CorrType = CORR_APPROX

     If (CorrType==CORR_APPROX) then
        Write(Log_Unit,*) 'SCF_LS: CorrType = CORR_APPROX'
     else
        Write(Log_Unit,*) 'SCF_LS: CorrType = CORR_BiCGSTABL'
     End If

    call flush_(Log_Unit)
     j = -1
     Do i=1, NumKpnts
        VecProcessed = 0
        Where (PsiInfo(:)%Kpnt == i) VecProcessed = 1
        j = MAX(j, SUM(VecProcessed))
     End Do
     
      
     If (2*j > ArraySize) then
        ArraySize = 2*j

        If (Associated(Hmat)) DeAllocate(Hmat, Omat, Eigen)
        Allocate(Hmat(ArraySize, Arraysize), Omat(ArraySize, ArraySize), &
             Vec(ArraySize, Arraysize), eigen(Arraysize))
     End If


     OldEnergy = CohesiveEnergy

     If (iter == 1) then
        Call CheckPotError( Log_Unit, StartPotError)
        IF (StartPotError == 0) StartPotError = 1E30
        Write(Log_Unit, *) 'AdaptiveMGP: CPOT Init Pot error:',StartPotError
     End If

     Call CleanUp_MemMap

    Call GetBuffer( Hx)
    Call GetBuffer( Ox)
    Call GetBuffer( PsiWork)
    Call GetBuffer( Ox2)

    Do Kpnt=1, NumKpnts
       Hmat = 0
       Omat = 0

       LUT_Count = 0
       LUT = 0


       !** Determine starting and ending cluster for K-point **
       Vec_toProcess = 0
       Where (PsiInfo(:)%Kpnt == Kpnt) Vec_toProcess = ClusterMap
       
       Cluster_End = MaxVal(Vec_toProcess)
       Where (Vec_toProcess == 0) Vec_toProcess = Cluster_End+1
       Cluster_Start = MinVal(Vec_toProcess)

       Vec_toProcess = MH_Skip
       Where (PsiInfo(:)%Kpnt == Kpnt) Vec_toProcess = MH_toProcess
    

       Call Phase_Generic( Vec_toProcess,  Kpnt)



       Psi_List = MH_Skip
       i = 0
       Do cluster=Cluster_Start, Cluster_End
          VecProcessed = MH_Skip
          Where (ClusterMap == cluster) VecProcessed = MH_toProcess
          VecSize = SUM(VecProcessed)

Write(Log_Unit,*) 'SCF_LS: Cluster = ',cluster, ' * VecSize=',VecSize
          Do c=1, VecSize
             i = i + 1

            k = 1
            Do While ((VecProcessed(k) == MH_processed) .AND. (k<=Mem_Mapsize))
               k = k + 1
            End do
            If (k > Mem_Mapsize) then
              Write(Error_Unit,*) 'Error in BlockDavidson',k,VecProcessed
              stop
            EndIf
            PsiIndex = k
            Call GetPsi_fromBuffer( PsiIndex, PsiWork)
            VecProcessed(k) = MH_Skip
            LUT(1,i) = PsiIndex
            LUT(2,PsiIndex) = i
            !write(Log_Unit,*)'LUT', i,PSiIndex

            Call Phase_Generic( Psi_List,  Kpnt)

            Call CalcHxAndOx( PsiWork, Hx, Ox, &
                 PsiInfo(PsiIndex)%PDOT, Kpnt, PsiInfo(PsiIndex)%PDOT_Stored)
          
            E3 = Basis_DotProd( PsiWork, Ox, .FALSE.)          
            E3 = SQRT(E3)
            PsiWork = PsiWork/E3
            Ox = Ox/E3
            Hx = Hx/E3
            Hmat(i, i) = Basis_DotProd( Hx, PsiWork, .FALSE.)
            Omat(i, i) = Basis_DotProd( Ox, PsiWork, .FALSE.)
            If (DoOrthog) Omat(i,i) = Basis_DotProd( Ox, PsiWork,.FALSE.)
            Call SavePsi_toBuffer( PsiIndex, PsiWork)


            If (SCF_DoGramSchmidt) then
               Call GetNextFreeIndex(j)
               PsiInfo(PsiIndex)%OxIndex = j
               PsiInfo(j)%Kpnt = BG_KPNT     ! special index for temp files
               Call SavePsi_toBuffer( j, Ox)
            End If

            !*** Calc the reduced H elements ***
            Do j=1, i-1
               Call GetNextPsi(Kindex, Psi2)
               k = LUT(2, Psi2%Index)    !** Get the LUT index
             !write(Log_Unit,*) 'matrix loop',j,k,Psi2%Index
               
               Term = Basis_DotProd( Hx, Psi2%Ptr, .FALSE.)
               Hmat(i, k) = Term
               Hmat(k, i) = CONJG(Term)
               !Write(*,*) 'GRRP: Inner Loop.  x2Hx1=',term
             
               Term = Basis_DotProd( Ox, Psi2%Ptr, .FALSE.)
               Omat(i, k) = Term


               Omat(K, i) = CONJG(Term)             
!Write(*,*) 'GRRP: Inner Loop.  x2Ox1=',term
!Write(Log_Unit,*) 'SCF_LS: GRRP: j=',j, ' * DI=',Psi2%Index, ' * i=',i
            End Do

       
            !*** Make the Correction vector ***

            !** Make sure the KE is stored 
            If (.NOT.PsiInfo(PsiIndex)%KE_Stored) then
               PsiInfo(PsiIndex)%KE=CalcKineticenergy(BZ%Ku(1:3,Kpnt),PsiWork)
               PsiInfo(PsiIndex)%KE_Stored = .TRUE.
            End If
             Call CalcCorr( PsiWork, Hx, Ox, &
                 PsiIndex, Kpnt, CorrType)

            !*** Now print the Error ***
            E1 = Basis_DotProd( PsiWork, Hx, .FALSE.)
            Hx = Hx - E1*Ox
            E2 = Basis_DotProd( Hx, Hx, .FALSE.)
            E3 = Basis_DotProd( PsiWork, Ox, .FALSE.)
            Write(Log_Unit,*) 'ERROR:',PsiIndex, E1, &
                 PsiInfo(PsiIndex)%Energy, E2, E3, Kpnt, &
                 PsiInfo(PsiIndex)%Occupancy

            Psi_List(PsiIndex) = MH_toProcess
         End do
       End Do

     
       !** Now add the correction vectors to the Matrix ***
       Vec_toProcess = MH_Skip
       VecProcessed = MH_Skip
      !!!! Lines replaced 4/28/01 NAWH
      !!!! VecSize = 0
      !!!! Do i=1, Mem_MapSize
      !!!!    If (PsiInfo(i)%Kpnt == Kpnt) then
      !!!!       VecSize = VecSize + 1
      !!!!       Vec_toProcess(i) = MH_toProcess
      !!!!       Vec_toProcess(PsiInfo(i)%CorrIndex) = MH_toProcess
      !!!!       VecProcessed(PsiInfo(i)%CorrIndex) = MH_toProcess
      !!!!    End If
      !!!! End Do

      VecSize=i
      Write(Log_Unit,*) 'Block Davidson -- Initial VecSize',VecSize
      Do i=1,VecSize
         j=LUT(1,i)
         Vec_toProcess(j) = MH_toProcess
         !write(Log_Unit,*)' i,j,corrindex',i,j,PsiInfo(j)%CorrIndex
         Vec_toProcess(PsiInfo(j)%CorrIndex) = MH_toProcess
         VecProcessed(PsiInfo(j)%CorrIndex) = MH_toProcess
      Enddo
   Psi_toProcess=VecProcessed

       Call Phase_Generic(  Vec_toProcess,  Kpnt)

Call PrintDate(Log_Unit,'SCF: Before Complete Hmat')
       n = VecSize
       Do i=1, VecSize
          n = n + 1
          Call GetNextPsi_UnProcessed(KIndex, VecProcessed,  &
               PsiIndex, PsiWork)

          If (KIndex == 0 ) then
             write(Error_Unit,*) 'Error in BD at GetNextPsi_UnProcessed', &
               ' VecSize,n = ', VecSize,n
             stop
          EndIf

          LUT(1,n) = PsiIndex
          LUT(2,PsiIndex) = n
            !write(Log_Unit,*)'LUT', n,PSiIndex

          If (SCF_DoGramSchmidt) then   !*** Orthog First if needed ***
             Call GetNextPsi(Kindex, Psi2)
             Do While (Kindex /= 0)
                k = PsiInfo(Psi2%Index)%OxIndex
                Call GetPsi_fromBuffer( k, Ox2)
                
                Term = Basis_DotProd( Ox2, PsiWork, .FALSE.)
                PsiWork = PsiWork - Psi2%Ptr * term

                Call GetNextPsi(Kindex, Psi2)
             End Do

             Call SavePsi_tobuffer( PsiIndex, PSiWork)
          End If


          PsiInfo(PsiIndex)%KE_Stored = .FALSE.
          PsiInfo(PsiIndex)%PDOT_Stored = .FALSE.
          Call CalcHxAndOx( PsiWork, Hx, Ox, &
               PsiInfo(PsiIndex)%PDOT, Kpnt, PsiInfo(PsiIndex)%PDOT_Stored)

          E3 = Basis_DotProd( PsiWork, Ox, .FALSE.)          
          E3 = SQRT(E3)
          PsiWork = PsiWork/E3
          Ox = Ox/E3
          Hx = Hx/E3
          Call SavePSi_toBuffer( PsiIndex, PsiWork)

          If (SCF_DoGramSchmidt) then
             Call GetNextFreeIndex(j)
             PsiInfo(PsiIndex)%OxIndex = j
             PsiInfo(j)%Kpnt = BG_KPNT
             Call SavePsi_toBuffer( j, Ox)
          End If
         
          Hmat(n, n) = Basis_DotProd( Hx, PsiWork, .FALSE.)
          Omat(n, n) = Basis_DotProd( Ox, PsiWork, .FALSE.)

!If (REAL(Omat(n,n))>1E-9) then          
!Write(Log_Unit,*) 'GRRP: Outer Loop.  xHx=',Hmat(n,n), ' * DI=',PsiIndex
          !*** Calc the reduced H elements ***
          Do j=1, n-1
             Call GetNextPsi(Kindex, Psi2)
             k = LUT(2, Psi2%Index)    !** Get the LUT index
        
           !write(Log_Unit,*) 'corr matrix loop',j,n,Psi2%Index
             Term = Basis_DotProd( Hx, Psi2%Ptr, .FALSE.)
             Hmat(n, k) = Term
             Hmat(k, n) = CONJG(Term)
!Write(Log_Unit,*) 'GRRP: Inner Loop.  x2Hx1=',term, ' * DI=',Psi2%Index

             Term = Basis_DotProd( Ox, Psi2%Ptr, .FALSE.)
             Omat(n, k) = Term
             Omat(k, n) = CONJG(Term)
!Write(*,*) 'GRRP: Inner Loop.  x2Ox1=',term
          End Do
       End Do
Call PrintDate(Log_Unit,'SCF: After Complete Hmat')

       !*** Now solve the system ***

       Call Diagonalizer(.NOT. SCF_DoGramSchmidt, n, ArraySize, NewSize, &
            Hmat, Omat, Eigen, Vec)

       Call Cluster_EigenValues(MinBands, VecSize, Eigen, NewSize)

!Write(Log_Unit,*) 'SCF_LS: VecSize=',VecSize, ' * n=',n, ' * LUT=',LUT(1,:)
!Write(Log_Unit,*) ' SCF_LS: Eigen=',Eigen(1:VecSize)

       Call MakeNewPsi( n, VecSize, Vec, Eigen, LUT, Kpnt)

       Call CleanUp_MemMap
       

!Write(Log_Unit, *) 'UpdateKpnt: After GRRP PsiInfo(:)%Kpnt=',PsiInfo(:)%Kpnt, ' * ClusterMap=',CLusterMap

       Where (PsiInfo(:)%Kpnt == Kpnt) 
          PsiInfo(:)%PDOT_Stored = .FALSE.       
          PsiInfo(:)%KE_Stored = .FALSE.
       End Where

    End Do  !** Kpnt

    Call FreeBuffer(Hx)
    Call FreeBuffer(Ox)
    Call FreeBuffer(PsiWork)
    Call FreeBuffer(Ox2)


    Call MakeBandMap
    Call CalcNewOccAndFermi
    Call ReCluster_All

    PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear PDOT's
    PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
    Call CalcMeritFunction( CohesiveEnergy)
    EnergyDif = ABS(CohesiveEnergy-OldEnergy)

    Call DisplayBufferInfo(Log_Unit)
    If (Atomic_Mode /= SIM_SEPM) then
       Call LRU_DisplayBufferInfo(LRU_Phase, Log_Unit)
       Call LRU_DisplayBufferInfo(LRU_Ylm, Log_Unit)
    End If
 If (.not.bandstructure_mode) then
    Write(Log_Unit,*) 'SCF_ITER: Current Iteration:',iter
    Write(Log_Unit,*) 'SCF: Cohesive Energy:',CohesiveEnergy, &
         'Old Energy: ',OldEnergy, '* Diff:',ABS(CohesiveEnergy-OldEnergy)
    Write(Log_Unit,*) 'SCF: New Fermi=',BZ%Fermi
    Call CheckPotError( Log_Unit, PotError)
 Else
    Write(Log_Unit,*) 'BAND_ITER: Current Iteration:',iter
    Write(Log_Unit,*) 'BAND: Cohesive Energy:',CohesiveEnergy, &
         'Old Energy: ',OldEnergy, '* Diff:',ABS(CohesiveEnergy-OldEnergy)
 EndIf
    Call DisplayEnergy(Log_Unit)

     Write(Log_Unit,*) 'ERROR: DI, CalcE, StoredE, Error, Mag, Kpnt, Occ'

     If (mod(iter,5) == 0 .AND. .not.bandstructure_mode) then
        Call StoreSolution_Binary( 'paw.wfn')
     End If
 
     If (.not.bandstructure_mode .AND. Best_Name /= "") then
        If (CohesiveEnergy > BestCohesiveEnergy) then
           BestCohesiveEnergy=CohesiveEnergy
           Call Store_BestPositions(Trim(Best_Name)//'.pos', &
                   BestCohesiveEnergy)
           Call StoreSolution_Binary(Trim(Best_Name)//'.wfn')
        EndIf
     EndIf

     Call Stop_Timer(Timer(Iter_Timer))
  End Do

  Converged = .FALSE.
  If (EnergyDif < EnergyTol) Converged = .TRUE.

  !*** Print Final Stats if converged ***
  If (Converged) then
    Do i=1,LastCLuster
        Call Phase_Cluster( i)
        Call CheckClusterError(Log_Unit, i, .TRUE.)
     End Do
  End If

  EnergyTol = EnergyDif

  Return
End Subroutine


!******************************************************************************
!
!  GenRandom - Generates a random Initial wave functions.
!
!******************************************************************************

Subroutine GenRandom

  Integer :: i, j, k, band, nili, Li, Mi, Orbitals_Size, Mesh_Size, psipos, tmp
  Real          :: Mesh_Step, Kvec(3), R(3), Occ, Energy, curr(2)
  Real, Pointer :: TPhi(:,:), Init_Occ(:), Glist(:,:)
  Integer, Pointer :: L_Values(:),  Valence_Orbitals(:)
  Integer  :: Psi_toUSe(Mem_MapSize), PSi_Kpnt(Mem_MapSize), DiskIndex
  Integer  :: NewPsi_Map(Mem_Mapsize), t, Kindex, G_Size
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  Complex, Pointer :: Psi(:), Work(:)
  Complex :: dot1, dot2, PDot(PLM_Max)
  Logical :: DoGramSchmidt
  Type(Mem_Handle), Pointer :: Psi2
  Real :: NewEigen_Map(Mem_MapSize), g1, g2

  DoGramSchmidt = .TRUE.

!RETURN !****DEBUG ONLY *******

  curr = (/V_NewMix, Dij_NewMix/)
  V_NewMix = 1;  Dij_NewMix = 1;
  
  band = 0

  If (Atomic_Mode /= SIM_SEPM) then
     !** First reset the mixing to NULL ***
     Call Anderson_ResetMix(SCFvalues%AC)
     Do i=1, Specific_Atoms
        Call Anderson_ResetMix(Atom_List(i)%Dij_AC)
     End Do
  End if

  !*** Make sure ALL wave functions are free ***
  PsiInfo(:)%Kpnt = 0
  PsiInfo(:)%DoSave = 0
  PsiInfo(:)%Available = Mem_Available
  PsiInfo(:)%MemBufIndex = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.
  PsiInfo(:)%KE_Stored = .FALSE.

  G_Size = FindGcut(LCAO_Gcut, Gpnt, Gpnt_Size(G_LOW))

  Write(Log_Unit,*) 'GenRandom: G_Size=',G_Size

  PsiInfo(:)%CorrIndex = 0

  Call GetBuffer( Psi)
  Call GetBuffer( Work)

  LastCLuster = 1

  Psi_toUSe = MH_Skip

  Allocate(Glist(4, 2*G_Size-1))

  Do k=1, NumKpnts
     write(Log_Unit,*) 'GenRandom: Processing Kpnt=',k
     kvec = BZ%Ku(:,k)

     Call Phase_Generic( Psi_toUse,  k)

     Band = 0
     Do i=1, MinBands + 1
        Call GetNExtFreeIndex(DiskIndex)
        PSi_toUse(DiskIndex) = MH_toProcess
        
        PsiInfo(DiskIndex)%Kpnt = k
        PsiInfo(DiskIndex)%Energy = 1
        PsiInfo(DiskIndex)%DoSave=1
        PsiInfo(DiskIndex)%Available=Mem_Used
        Occ = TotalElectrons/(MinBands+1.0)*BZ%WtUniq(1,k)
        PsiInfo(DiskIndex)%Occupancy  = Occ 
        ClusterMap(DiskIndex) = 1
        ClusterSize(k) = ClusterSize(k) + 1

        Psi = 0
        Do j=1, Gall_size(G_LOW)
           Call Random_Number(g1)
           Call Random_Number(g2)
           psi(j) = CMPLX(g1,g2)
        End Do

        Psi = Psi / SQRT(REAL(DOT_PRODUCT(PSi,Psi)))
        Psi(k) = 1
!-------------                            
        If (DoGramSchmidt) then   !*** Orthog First if needed ***
           Psi_Kpnt = MH_Skip
           Where (PsiInfo(:)%Kpnt == k) Psi_Kpnt = MH_toProcess
           Psi_Kpnt(DiskIndex) = MH_Skip
           Call Phase_Generic(Psi_Kpnt,  k)
           
           Call GetNextPsi(Kindex, Psi2)
           Do While (Kindex /= 0)
              t = PsiInfo(Psi2%Index)%OxIndex
              Call GetPsi_fromBuffer( t, Work)
              
              dot1 = Basis_DotProd( Work, Psi, .FALSE.)
              Psi = Psi - Psi2%Ptr * dot1
              
              Call GetNextPsi(Kindex, Psi2)
           End Do
        End If
!--------------

        Call FilterValue(Psi)
        PsiInfo(DiskIndex)%PDOT_Stored = .FALSE.
        Call Opsi( Psi, Work, &
             PDOT, k, PsiInfo(DiskIndex)%PDOT_Stored)
        Call FilterValue(Work)
        
        dot1 = Basis_DotProd( Psi, Work, .FALSE.)
! Write(Log_Unit,*) 'GenRandom: DI=',DiskIndex, ' * DOT1=',dot1
! Write(*,*) 'GenRandom: DI=',DiskIndex, ' * DOT1=',dot1
        dot1 = SQRT(Real(Dot1))
if (real(dot1) < 1E-20) then
   Write(*,*) 'GenRandom: DI=',DiskIndex, ' * DOT1=',dot1
   STOP
end IF
        Psi = Psi / Dot1
        Work = Work / Dot1
        PsiInfo(DiskIndex)%PDOT = PsiInfo(DiskIndex)%PDOT / dot1
!Write(Log_Unit,*) 'GenRandom: DI=',DiskIndex, ' * DOT1=',dot1
        
        Call SavePsi_toBuffer( DiskIndex, Psi)
 
        If (DoGramSchmidt) then
           Call GetNextFreeIndex(t)
           PsiInfo(DiskIndex)%OxIndex = t
           PsiInfo(t)%Kpnt = BG_Kpnt
           Call SavePsi_toBuffer( t, Work)
        End If
     End Do

     Call CleanUp_MemMap

  End Do

  DeAllocate(GList)

  Call FreeBuffer(PSi)
  Call FreeBuffer(Work)


  Call MakeBandMap
  Occupancy = 0.0
  Do i=1, Mem_MapSize
     Band = BandMap(1,i)
     K = BandMap(2,i)
   
     If (K /= 0) then
        Occupancy(Band, K) = PsiInfo(i)%Occupancy
     End If
  End Do


  Call CalcMeritFunction( Energy)

  Write(Log_Unit,*) 'GenRandom: Initial Wave functions'
  Do j=1, NumKpnts
     PSi_Kpnt = MH_Skip
     Where (PsiInfo(:)%Kpnt == j) PSi_Kpnt = MH_toProcess

!**     Where (PsiInfo(:)%Kpnt == j) PsiInfo(:)%PDOT_STORED = .FALSE.

     Call Phase_Generic( Psi_Kpnt, j)
     k = 0
     Do i=1, Mem_MapSize
        If (PsiInfo(i)%Kpnt == j) then
           k = k + 1
        End IF
     End Do

     NewPsi_Map = 0; NewEigen_Map = 0
     Call UpdateKpnt( j, k, Psi_Kpnt)
  End Do

  PsiInfo(:)%PDOT_Stored = .FALSE.   !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.   !** and KE's

!--------------



  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_All

  Call CalcMeritFunction( Energy)

  V_NewMix = Curr(1);  Dij_NewMix = Curr(2)

  Return
End Subroutine

!******************************************************************************
!
!  GenLCAO - Generates all the Initial wave LCAO wave functions and inital
!            occupancies.
!
!******************************************************************************

Subroutine GenLCAO

  Integer :: i, j, k, band, nili, Li, Mi, Orbitals_Size, Mesh_Size, psipos, tmp
  Real          :: Mesh_Step, Kvec(3), R(3), Occ, Energy, curr(2)
  Real, Pointer :: TPhi(:,:), Init_Occ(:), Glist(:,:)
  Integer, Pointer :: L_Values(:),  Valence_Orbitals(:)
  Integer  :: Psi_toUSe(Mem_MapSize), PSi_Kpnt(Mem_MapSize), DiskIndex
  Integer  :: NewPsi_Map(Mem_Mapsize), t, Kindex, G_Size
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  Complex, Pointer :: Psi(:), Work(:)
  Complex :: dot1, dot2, PDot(PLM_Max)
  Logical :: DoGramSchmidt
  Type(Mem_Handle), Pointer :: Psi2
  Real :: NewEigen_Map(Mem_MapSize)

  DoGramSchmidt = .FALSE.

!RETURN !****DEBUG ONLY *******

  curr = (/V_NewMix, Dij_newmix/)
  V_NewMix = 1;  Dij_NewMix = 1;
  
  band = 0
  
  !** First reset the mixing to NULL ***
  Call Anderson_ResetMix(SCFvalues%AC)
  Do i=1, Specific_Atoms
     Call Anderson_ResetMix(Atom_List(i)%Dij_AC)
  End Do

  !*** Make sure ALL wave functions are free ***
  PsiInfo(:)%Kpnt = 0
  PsiInfo(:)%DoSave=0
  PsiInfo(:)%Available=Mem_Available
  PsiInfo(:)%MemBufIndex = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.
  PsiInfo(:)%KE_Stored = .FALSE.

  G_Size =  Gpnt_Size(G_LOW)

  Write(Log_Unit,*) 'GenLCAO: G_Size=',G_Size

  PsiInfo(:)%CorrIndex = 0

  Call GetBuffer( Psi)
  Call GetBuffer( Work)

  LastCLuster = 1

  Psi_toUSe = MH_Skip

  Allocate(Glist(4, 2*G_Size-1))

  Do k=1, NumKpnts
!!     write(Log_Unit,*) 'GenLCAO: Processing Kpnt=',k
     kvec = BZ%Ku(:,k)

     i = 2*G_Size-1
     Call AllocLCAO(i)
     Call CalcRadG_LCAO(G_size, Kvec, Glist)

     Call Phase_Generic( Psi_toUse,  k)

     Band = 0
     Do i=1, Specific_Atoms
        A => Atom_List(i)
        AT => AtomType_Info(A%TypeIndex)
        L_Values => AT%L_Value
        Orbitals_Size = A%Orbitals_Size
        Valence_Orbitals => A%VAlence_Orbitals
        MEsh_Step = AT%LCAO_Step
        Mesh_Size = AT%LCAO_Size
        Tphi => AT%RadG_LCAO
        Init_Occ => A%Init_Occ
        R = A%Pos

        Do j=1, Orbitals_Size
           nili = Valence_Orbitals(j)
           Li = L_Values(nili)
           Occ = Init_Occ(j)/(2*Li+1)*BZ%WtUniq(1,k)

           Do mi = -Li, Li
              Call GetNExtFreeIndex(DiskIndex)
              PSi_toUse(DiskIndex) = MH_toProcess

              PsiInfo(DiskIndex)%Kpnt = k
              PsiInfo(DiskIndex)%DoSave = 1
              PsiInfo(DiskIndex)%Available = Mem_Used
              PsiInfo(DiskIndex)%Energy = Li
              PsiInfo(DiskIndex)%Occupancy = Occ
              ClusterMap(DiskIndex) = 1
              ClusterSize(k) = ClusterSize(k) + 1

              Call CalcLCAO( Psi, Tphi(:,nili), Mesh_Size, &
                   Mesh_Step, Li, Mi, Kvec, R, LCAO_Cells, GList, G_Size)

!-------------                            
              If (DoGramSchmidt) then   !*** Orthog First if needed ***
                 Psi_Kpnt = MH_Skip
                 Where (PsiInfo(:)%Kpnt == k) Psi_Kpnt = MH_toProcess
                 Psi_Kpnt(DiskIndex) = MH_Skip
                 Call Phase_Generic(Psi_Kpnt,  k)

                 Call GetNextPsi(Kindex, Psi2)
                 Do While (Kindex /= 0)
                    t = PsiInfo(Psi2%Index)%OxIndex
                    
                    Call GetPsi_fromBuffer( t, Work)
                    
                    dot1 = Basis_DotProd( Work, Psi, .FALSE.)
                    Psi = Psi - Psi2%Ptr * dot1
                    
                    Call GetNextPsi(Kindex, Psi2)
                 End Do
              End If
!--------------

              PsiInfo(DiskIndex)%PDOT_Stored = .FALSE.

              Call FilterValue(Psi)
              Call Opsi( Psi, Work, PDot, k, &
                   PsiInfo(DiskIndex)%PDOT_Stored)
              Call FilterValue(Work)
              
              dot1 = Basis_DotProd( Psi, Work, .FALSE.)
! Write(Log_Unit,*) 'GenLCAO: DI=',DiskIndex, ' * DOT1=',dot1
              dot1 = SQRT(Real(Dot1))
              Psi = Psi / Dot1
              Work = Work / Dot1
              PsiInfo(DiskIndex)%PDOT = PDOT / dot1
! Write(Log_Unit,*) 'GenLCAO: DI=',DiskIndex, ' * DOT1=',dot1

              Call SavePsi_toBuffer( DiskIndex, Psi)
 
              If (DoGramSchmidt) then
                 Call GetNextFreeIndex(t)
                 PsiInfo(DiskIndex)%OxIndex = t
                 PsiInfo(t)%Kpnt = BG_Kpnt
                 Call SavePsi_toBuffer( t, Work)
              End If
           End Do
        End Do

        Call CleanUp_MemMap

     End Do

     Call FreeLCAO

  End Do

  DeAllocate(GList)

  Call FreeBuffer(PSi)
  Call FreeBuffer(Work)


  Call MakeBandMap
  Occupancy = 0.0
  Do i=1, Mem_MapSize
     Band = BandMap(1,i)
     K = BandMap(2,i)
   
     If (K /= 0) then
        Occupancy(Band, K) = PsiInfo(i)%Occupancy
     End If
  End Do


  Call CalcMeritFunction(Energy)

  Write(Log_Unit,*) 'GenLCAO: Initial Wave functions'
  Do j=1, NumKpnts
     PSi_Kpnt = MH_Skip
     Where (PsiInfo(:)%Kpnt == j) PSi_Kpnt = MH_toProcess

     Where (PsiInfo(:)%Kpnt == j) PsiInfo(:)%PDOT_STORED = .FALSE.
     Where (PsiInfo(:)%Kpnt == j) PsiInfo(:)%KE_STORED = .FALSE.

     Call Phase_Generic( Psi_Kpnt,  j)
     k = 0
     Do i=1, Mem_MapSize
        If (PsiInfo(i)%Kpnt == j) then
           Call CheckError(i, Log_Unit)
           k = k + 1
        End IF
     End Do

     Call UpdateKpnt( j, k, Psi_Kpnt)
  End Do

  PsiInfo(:)%PDOT_Stored = .FALSE.   !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.   !** and KE's

!--------------

!write(Log_unit,*) 'before recluster in genLCAO'

  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_All

!write(Log_unit,*) 'before merit function in genLCAO'
  Call CalcMeritFunction( Energy)


  V_NewMix = Curr(1);  Dij_newmix = Curr(2)

  Return
End Subroutine


!******************************************************************************
!
!  SCF_Solver - Solves the SCF Eigenvalue/Potential problem.
!           This is the routine called bythe user.
!
!  MaxIter   - Max number of SCF steps
!  EnergyTol - Energy Convergence tolerance upon input and on output
!              the energy error
!  Iter      - Number of iterations performed
!
!******************************************************************************

Subroutine SCF_Solver( MaxIter, EnergyTol, iter)
  Integer,          Intent(IN)    :: MaxIter
  Real,     Intent(INOUT) :: EnergyTol
  Integer,          Intent(OUT)   :: Iter

  Integer      :: NumPsi, i,j, MaxAMG,  PM_Slice(3), PM(3)
  Integer      :: UsedMap(Mem_MapSize), Band, Kpnt, TotalIter
  Logical      :: Converged, DoTest
  Real :: ETol, EDif, Eold, Enew, etemp
  Character*100 :: ext
  Logical :: DoHamCheck
  Integer :: L, M, Li, Mi, Lj, Mj, VecSize, Scan(Mem_MapSize)
  Integer, Pointer :: LUT(:)
  character*100 :: Fname1, fname2
  Complex, Pointer :: Work1(:), Work2(:), Work3(:), Ptr(:)
  Logical :: DoOrthog

  !Global_Start_Time = MPI_WTIME()
  !Start_Time = MPI_WTIME()
  !Parallel_Time = 0
  Serial_Time = 0


!*RETURN  !**FIXME**

  Call PrintDate(Log_Unit, 'SimNLFMG: Start of Routine!')

  BG_Mode = 0

  If (DoLCAO /= 0) then
     ClusterSize = 0       !** Init cluster information
     LastCluster = 0
     ClusterMap = 0

     If (DoLCAO == GUESS_LCAO) then
        Call GenLCAO
     else
        Call GenRandom
     End If
  else   !** Already loaded an initial guess from disk ***
  End If

!**  Prev_CohesiveEnergy = 1E30
 
  Write(Log_Unit,*) 'SCF_Solver: LOOPing started!'
  Call PrintDate(Log_Unit, 'SCF_START:')

  TotalIter = 0
  Converged = .FALSE.

  Etol = EnergyTol

     Call BlockDavidson ( MaxIter, ETol, Converged, TotalIter)


  Call PrintDate(Log_Unit, 'SCF: End of Routine!')
  EnergyTol = Etol

  !Parallel_Time = Parallel_Time + MPI_WTIME() - Start_Time
  !edif = MPI_WTIME() - Global_Start_Time
    
  Write(Log_Unit,*) 'SCF:      Serial Time :',Serial_Time
  Write(Log_Unit,*) 'SCF:    Parallel Time :',Parallel_Time
  Write(Log_Unit,*) 'SCF:  Serial+Parallel :',Serial_Time+Parallel_Time
  Write(Log_Unit,*) 'SCF:      Global Time :',edif

  iter = TotalIter !** Return total number of iterations ***

  If(.not.bandstructure_mode) &
        Call StoreSolution_Binary('paw.wfn')  !** Store solution **

  Return
End Subroutine


!******************************************************************************
!
!  InitSolver - Initializes the Solver module for use
!
!******************************************************************************

Subroutine InitSolver
  Integer :: i, j, k, err
  Character*100 :: msg
  Real :: t1
  Complex :: c


  BaseFree = MemFree

  Write(Log_Unit,*) 'InitSolver: After Alloc MemFree=',MemFree, &
       ' * BaseFree=',BaseFree


  Return
End Subroutine

End Module
